"""
IOD Module – Patient Study

C.7.2.2
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.2.2.html
"""

from nii2dcm.module import Module


class PatientStudy(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'PatientStudy'

        self.ds.PatientAge = ""
        self.ds.PatientWeight = ""



"""
IOD Module – VOI LUT

C.11.2
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.11.2.html
"""

from nii2dcm.module import Module


class VOILUT(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'VOILUT'

        # WindowCenter
        # From NEMA: Required if VOI LUT Sequence (0028,3010) is not present. May be present otherwise.
        self.ds.WindowCenter = ''
    
        # WindowWidth
        # Required if Window Center (0028,1050) is present.
        self.ds.WindowWidth = ''



"""
IOD Module – Patient

C.7.1.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.html#sect_C.7.1.1
"""

from nii2dcm.module import Module


class Patient(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'PatientModule'

        self.ds.PatientName = 'Lastname^Firstname'
        self.ds.PatientID = '12345678'
        self.ds.PatientSex = ''
        self.ds.PatientBirthDate = ''



"""
IOD Module – General Image

C.7.6.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.6.html#sect_C.7.6.1
"""

from nii2dcm.module import Module


class GeneralImage(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'GeneralImage'

        self.ds.InstanceNumber = ''
        self.ds.PatientOrientation = ''
        self.ds.ContentDate = ''
        self.ds.ContentTime = ''
        self.ds.ImageType = ['SECONDARY', 'DERIVED']

        # LossyImageCompression
        # Enumerated values either:
        # 00 = no lossy compression
        # 01 = lossy compression
        self.ds.LossyImageCompression = '00'




"""
Classes for manipulating NIfTI files

Tom Roberts
"""

import numpy as np


class Nifti:

    def get_nii2dcm_parameters(nib_nii):
        """
        Get general NIfTI header parameters relevant for DICOM tag transferal.
        :nib_nii - NIfTI loaded with nibabel
        :nii_parameters - parameters to transfer to DICOM header
        """

        def fnT1N(A, N):
            # Subfn: calculate T1N vector
            # A = affine matrix [4x4]
            # N = slice number (counting from 1)
            T1N = A.dot([0, 0, N - 1, 1])
            return T1N

        # load nifti
        nii_img = nib_nii.get_fdata()

        # volume dimensions
        if nib_nii.header['dim'][4] == 1:
            nX, nY, nZ, nF = nib_nii.header['dim'][1], nib_nii.header['dim'][2], nib_nii.header['dim'][3], 1
            dimX, dimY, dimZ = nib_nii.header['pixdim'][1], nib_nii.header['pixdim'][2], nib_nii.header['pixdim'][3]

        elif nib_nii.header['dim'][4] > 1:
            print("Warning: Nifti is not 3-dimensional.")

        # Instances & Slice Spacing
        nInstances = nZ*nF
        sliceIndices = np.repeat(range(1, nZ+1), nF)
        voxelSpacing = dimZ
        zLocLast = (voxelSpacing * nZ) - voxelSpacing
        sliceLoca = np.repeat( np.linspace(0, zLocLast, num=nZ), nF)

        # Windowing & Signal Intensity
        maxI = np.amax(nii_img)
        minI = np.amin(nii_img)
        windowCenter = round((maxI - minI) / 2)
        windowWidth = round(maxI - minI)
        rescaleIntercept = 0
        rescaleSlope = 1

        # FOV
        fovX = nX * dimX
        fovY = nY * dimY
        fovZ = nZ * dimZ

        # slice positioning in 3-D space
        # nb: -1 for dir cosines gives consistent orientation between Nifti and DICOM in ITK-Snap
        A = nib_nii.affine
        dircosX = -1 * A[:3, 0] / dimX
        dircosY = -1 * A[:3, 1] / dimY

        image_pos_patient_array = []
        for iInstance in range(0, nInstances):
            T1N = fnT1N(A, iInstance)
            image_pos_patient_array.append([T1N[0], T1N[1], T1N[2]])

        # output dictionary
        nii2dcm_parameters = {

            # series parameters
            'dimX': dimX,
            'dimY': dimY,
            'SliceThickness': str(dimZ),
            'SpacingBetweenSlices': str(dimZ),
            'AcquisitionMatrix': [0, nX, nY, 0],
            'Rows': nX,
            'Columns': nY,
            'NumberOfSlices': nZ,
            'NumberOfInstances': nZ*nF,
            'PixelSpacing': [dimX, dimY],
            'FOV': [fovX, fovY, fovZ],
            'SmallestImagePixelValue': minI,
            'LargestImagePixelValue': maxI,
            'WindowCenter': str(windowCenter),
            'WindowWidth': str(windowWidth),
            'RescaleIntercept': str(rescaleIntercept),
            'RescaleSlope': str(rescaleSlope),
            'SpacingBetweenSlices': round(float(dimZ), 2),
            'ImageOrientationPatient': [dircosY[0], dircosY[1], dircosY[2], dircosX[0], dircosX[1], dircosX[2]],
            # alternative:
            # 'ImageOrientationPatient': [dircosX[0], dircosX[1], dircosX[2], dircosY[0], dircosY[1], dircosY[2]],

            # instance parameters
            'InstanceNumber': sliceIndices,
            'SliceLocation': sliceLoca,
            'ImagePositionPatient': image_pos_patient_array

        }

        return nii2dcm_parameters



"""
IOD Module – Frame of Reference

C.7.4.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.4.html#sect_C.7.4.1
"""

from nii2dcm.module import Module


class FrameOfReference(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'FrameOfReference'

        self.ds.FrameOfReferenceUID = ''
        # self.ds.PositionReferenceIndicator = ''  # TODO add robustly



"""
IOD Module – General Reference

C.12.4
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.12.4.html
"""

from nii2dcm.module import Module


class GeneralReference(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'GeneralReference'

        # TODO update placeholder class
        pass




"""
module.py
"""

from pydicom.dataset import Dataset


class Module:

    def __init__(self):
        """
        Creates generic Module object with empty Pydicom Dataset object

        nb: below somewhat superfluous – building out for future
        """
        self.module_type = 'GenericModule'

        self.ds = Dataset()



"""
IOD Module – Image Pixel

C.7.6.3
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.6.3.html

and

C.7-11c
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.6.3.3.html#table_C.7-11c
"""


from nii2dcm.module import Module


class ImagePixel(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'ImagePixel'

        self.ds.Rows = ''
        self.ds.Columns = ''

        # BitsAllocated, BitsStored, HighBit
        # Setting default = 1 for purposes of basic DICOM creation. Should be overwritten by Dicom subclass.
        self.ds.BitsAllocated = 1
        self.ds.BitsStored = ''
        self.ds.HighBit = ''
    
        # PixelRepresentation
        # Enumerated values either: unsigned integer or two's complement
        # Setting = 0, as observed in real DICOM
        self.ds.PixelRepresentation = 0
    
        self.ds.SmallestImagePixelValue = ''
        self.ds.LargestImagePixelValue = ''
    
        # PixelData written in dcm_writer via Pydicom
        self.ds.PixelData = ''



"""
IOD Module – General Acquisition

C.7.10.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.10.html#sect_C.7.10.1
"""

from nii2dcm.module import Module


class GeneralAcquisition(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'GeneralAcquisition'

        # TODO generate Acquisition values here - perhaps with date/time subfunction - or perform within Dicom class
        #  (as currently doing)?
        self.ds.AcquisitionNumber = ''
        self.ds.AcquisitionDate = ''
        self.ds.AcquisitionTime = ''



"""
Utility functions for nii2dcm
"""

# TODO(tomaroberts) create DicomDict class/functions which enable user to add custom attributes to nii2dcm Dicom
#  subclasses. Code below is currently non-functioning, but retained as reminder for future development.

from typing import Dict
from pydicom.datadict import DicomDictionary, keyword_dict


def dcm_dictionary_update(new_dict_items: Dict):
    """
    Update Pydicom DicomDictionary object with non-standard Private tags. Note: these tags are not added to the
    instantiated nii2dcm Dicom object; the DicomDictionary object is just updated meaning that these tags are now
    available to be written if the user desires.

    Tuple definitions: (VR, VM, description, is_retired flag, keyword)

    See: https://pydicom.github.io/pydicom/stable/auto_examples/metadata_processing/plot_add_dict_entries.html
    """

    # TODO temporary pass – implement functionality and remove later
    pass

    # Update the dictionary itself
    DicomDictionary.update(new_dict_items)

    # Update the reverse mapping from name to tag
    new_names_dict = dict([(val[4], tag) for tag, val in new_dict_items.items()])
    keyword_dict.update(new_names_dict)


# Example dictionary
# Supply as: dcm_dictionary_update(new_dict_items_PRIDESVR)
new_dict_items_PRIDESVR = {
    0x20011002: ('IS', '1', "Chemical Shift Number MR", '', 'ChemicalShiftNumberMR'),
    0x20011008: ('IS', '1', "Phase Number", '', 'PhaseNumber'),
    0x2001100a: ('IS', '1', "Slice Number MR", '', 'SliceNumberMR'),
    0x2001100b: ('CS', '1', "Slice Orientation", '', 'SliceOrientation'),
    0x20011014: ('SL', '1', "Number Of Echoes", '', 'NumberOfEchoes'),
    0x20011015: ('SS', '1', "Number Of Locations", '', 'NumberOfLocations'),
    0x20011016: ('SS', '1', "Number Of PC Directions", '', 'NumberOfPCDirections'),
    0x20011017: ('SL', '1', "Number Of Phases MR", '', 'NumberOfPhasesMR'),
    0x20011018: ('SL', '1', "Number Of Slices MR", '', 'NumberOfSlicesMR'),
    0x20011020: ('LO', '1', "Scanning Technique", '', 'ScanningTechnique'),
    0x20011025: ('SH', '1', "Echo Time Display MR", '', 'EchoTimeDisplayMR'),
    0x20011060: ('SL', '1', "Number Of Stacks", '', 'NumberOfStacks'),
    0x20011063: ('CS', '1', "Examination Source", '', 'ExaminationSource'),
    # 0x2001107b: ('IS', '1', "Acquisition Number", '', 'AcquisitionNumber'), # Philips Private alternative
    0x20011081: ('IS', '1', "Number Of Dynamic Scans", '', 'NumberOfDynamicScans'),
    0x2001101a: ('FL', '3', "PC Velocity", '', 'PCVelocity'),
    0x2001101d: ('IS', '1', "Reconstruction Number MR", '', 'ReconstructionNumberMR'),
    0x20051035: ('CS', '1', '', '', 'unknowntag20051035'),
    # PIXEL --- this seems to also correspond to MRSeriesDataType?
    0x20051011: ('CS', '1', 'MR Image Type', '', 'MRImageType'),
    0x2005106e: ('CS', '1', 'MR Scan Sequence', '', 'MRScanSequence'),

    # Philips "Stack" Tags
    0x2001105f: ('SQ', '1', 'Stack', '', 'Stack'),
    0x20010010: ('LO', '1', "Private Creator", '', 'PrivateCreator20010010'),
    0x2001102d: ('SS', '1', 'StackNumberOfSlices', '', 'StackNumberOfSlices'),
    0x20011032: ('FL', '1', 'StackRadialAngle', '', 'StackRadialAngle'),
    0x20011033: ('CS', '1', 'StackRadialAxis', '', 'StackRadialAxis'),
    0x20011035: ('SS', '1', 'MRSeriesDataType', '', 'MRSeriesDataType'),  # SS - StackSliceNumber ?
    0x20011036: ('CS', '1', 'StackType', '', 'StackType'),
    0x20050010: ('LO', '1', 'Private Creator', '', 'PrivateCreator20050010'),
    # Is this duplicate necessary with entry above?
    0x20050011: ('LO', '1', 'Private Creator', '', 'PrivateCreator20050011'),
    0x20050012: ('LO', '1', 'Private Creator', '', 'PrivateCreator20050012'),
    0x20050013: ('LO', '1', 'Private Creator', '', 'PrivateCreator20050013'),
    0x20050014: ('LO', '1', 'Private Creator', '', 'PrivateCreator20050014'),
    0x20050015: ('LO', '1', 'Private Creator', '', 'PrivateCreator20050015'),
    0x20051071: ('FL', '1', 'MRStackAngulationAP', '', 'MRStackAngulationAP'),
    0x20051072: ('FL', '1', 'MRStackAngulationFH', '', 'MRStackAngulationFH'),
    0x20051073: ('FL', '1', 'MRStackAngulationRL', '', 'MRStackAngulationRL'),
    0x20051074: ('FL', '1', 'MRStackFovAP', '', 'MRStackFovAP'),
    0x20051075: ('FL', '1', 'MRStackFovFH', '', 'MRStackFovFH'),
    0x20051076: ('FL', '1', 'MRStackFovRL', '', 'MRStackFovRL'),
    0x20051078: ('FL', '1', 'MRStackOffcentreAP', '', 'MRStackOffcentreAP'),
    0x20051079: ('FL', '1', 'MRStackOffcentreFH', '', 'MRStackOffcentreFH'),
    0x2005107a: ('FL', '1', 'MRStackOffcentreRL', '', 'MRStackOffcentreRL'),
    0x2005107b: ('CS', '1', 'MRStackPreparationDirection', '', 'MRStackPreparationDirection'),
    0x2005107e: ('FL', '1', 'MRStackSliceDistance', '', 'MRStackSliceDistance'),
    0x20051081: ('CS', '1', 'MRStackViewAxis', '', 'MRStackViewAxis'),
    0x2005143c: ('FL', '1', 'MRStackTablePosLong', '', 'MRStackTablePosLong'),
    0x2005143d: ('FL', '1', 'MRStackTablePosLat', '', 'MRStackTablePosLat'),
    0x2005143e: ('FL', '1', 'MRStackPosteriorCoilPos', '', 'MRStackPosteriorCoilPos'),
    0x20051567: ('IS', '1', 'MRPhilipsX1', '', 'MRPhilipsX1'),

    # Phase Contrast/Velocity Tags
    0x00089209: ('CS', '1', "Acquisition Contrast", '', 'AcquisitionContrast'),
    0x00189014: ('CS', '1', "Phase Contrast", '', 'PhaseContrast'),
    0x00189090: ('FD', '3', "Velocity Encoding Direction", '', 'VelocityEncodingDirection'),
    0x00189091: ('FD', '1', "Velocity Encoding Minimum Value", '', 'VelocityEncodingMinimumValue'),
}



"""
IOD Module – General Series

C.7.3.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.3.html#sect_C.7.3.1
"""

from nii2dcm.module import Module


class GeneralSeries(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'GeneralSeries'

        self.ds.Modality = ''  # initiated, should be defined by Dicom subclass
        self.ds.SeriesInstanceUID = ''  # initiated, but value set by Dicom.init_series_tags()
        self.ds.SeriesNumber = ''  # initiated, but value set by Dicom.init_series_tags()
        self.ds.ProtocolName = "nii2dcm_DICOM"

        # PatientPosition
        # From NEMA: required for images where Patient Orientation Code Sequence (0054,0410) is not present and whose
        # SOP Class is one of the following:
        # CT ("1.2.840.10008.5.1.4.1.1.2") or
        # MR ("1.2.840.10008.5.1.4.1.1.4") or
        # Enhanced CT ("1.2.840.10008.5.1.4.1.1.2.1") or
        # Enhanced MR Image ("1.2.840.10008.5.1.4.1.1.4.1") or
        # Enhanced Color MR Image ("1.2.840.10008.5.1.4.1.1.4.3") or
        # MR Spectroscopy ("1.2.840.10008.5.1.4.1.1.4.2")
        self.ds.PatientPosition = ''

        self.ds.AccessionNumber = "ABCXYZ"



"""
IOD Module – General Study

C.7.2.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.2.html#sect_C.7.2.1
"""


from nii2dcm.module import Module


class GeneralStudy(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'GeneralStudy'

        self.ds.StudyInstanceUID = ''  # attribute initialised but value initialised in Dicom class
        self.ds.StudyDescription = ''
        self.ds.ReferringPhysicianName = ''





"""
IOD Module – Image Plane

C.7.6.2
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.6.2.html
"""

from nii2dcm.module import Module


class ImagePlane(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'ImagePlane'

        self.ds.PixelSpacing = ''
        self.ds.ImageOrientationPatient = ''  # TODO set here or in Dicom class?
        self.ds.ImagePositionPatient = ''  # TODO set here or in Dicom class?
        self.ds.SliceThickness = ''
        self.ds.SpacingBetweenSlices = ''
        self.ds.SliceLocation = ''  # maths: https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.6.2.html#sect_C.7.6.2.1.2



"""
nii2dcm runner
"""
from os.path import abspath

import nibabel as nib
import pydicom as pyd

import nii2dcm.nii
import nii2dcm.svr
from nii2dcm.dcm_writer import (
    transfer_nii_hdr_series_tags,
    transfer_nii_hdr_instance_tags,
    transfer_ref_dicom_series_tags,
    write_slice
)


def run_nii2dcm(input_nii_path, output_dcm_path, dicom_type=None, ref_dicom_file=None):
    """
    Execute NIfTI to DICOM conversion

    :param input_nii_path: input .nii/.nii.gz file
    :param output_dcm_path: output DICOM directory
    :param dicom_type: specified by user on command-line
    :param ref_dicom: reference DICOM file for transferring Attributes
    """

    # load NIfTI
    nii = nib.load(input_nii_path)

    # get pixel data from NIfTI
    # TODO: create method in nii class
    nii_img = nii.get_fdata()
    nii_img = nii_img.astype("uint16")  # match DICOM datatype

    # get NIfTI parameters
    nii2dcm_parameters = nii2dcm.nii.Nifti.get_nii2dcm_parameters(nii)

    # initialise nii2dcm.dcm object
    # --dicom_type specified on command line
    if dicom_type is None:
        dicom = nii2dcm.dcm.Dicom('nii2dcm_dicom.dcm')

    if dicom_type is not None and dicom_type.upper() in ['MR', 'MRI']:
        dicom = nii2dcm.dcm.DicomMRI('nii2dcm_dicom_mri.dcm')

    if dicom_type is not None and dicom_type.upper() in ['SVR']:
        dicom = nii2dcm.svr.DicomMRISVR('nii2dcm_dicom_mri_svr.dcm')
        nii_img = nii.get_fdata()
        nii_img[nii_img < 0] = 0  # set background pixels = 0 (negative in SVRTK)
        nii_img = nii_img.astype("uint16")

    # load reference DICOM object
    # --ref_dicom_file specified on command line
    if ref_dicom_file is not None:
        ref_dicom = pyd.dcmread(ref_dicom_file)

    # transfer Series tags from NIfTI
    transfer_nii_hdr_series_tags(dicom, nii2dcm_parameters)

    # transfer tags from reference DICOM
    # IMPORTANT: this deliberately happens last in the DICOM tag manipulation process so that any tag values transferred
    # from the reference DICOM override any values initialised by nii2dcm
    if ref_dicom_file is not None:
        transfer_ref_dicom_series_tags(dicom, ref_dicom)

    """
    Write DICOM files
    - Transfer NIfTI parameters and write slices, instance-by-instance
    """
    print('nii2dcm: writing DICOM files ...')  # TODO use logger

    for instance_index in range(0, nii2dcm_parameters['NumberOfInstances']):

        # Transfer Instance tags
        transfer_nii_hdr_instance_tags(dicom, nii2dcm_parameters, instance_index)

        # Write slice
        write_slice(dicom, nii_img, instance_index, output_dcm_path)

    print(f'nii2dcm: DICOM files written to: {abspath(output_dcm_path)}')  # TODO use logger



"""
nii2dcm entrypoint code and command line interface (CLI)
"""

import sys
import argparse
from pathlib import Path
from nii2dcm.run import run_nii2dcm


def cli(args=None):
    """
    Run nii2dcm via command line
    """
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog="nii2dcm",
        description="nii2dcm - NIfTI file to DICOM conversion"
    )

    parser.add_argument("input_file", type=str, help="[.nii/.nii.gz] input NIfTI file")
    parser.add_argument("output_dir", type=str, help="[directory] output DICOM path")
    parser.add_argument("-d", "--dicom_type", type=str, help="[string] type of DICOM. e.g. MR, CT, US, XR, etc.")
    parser.add_argument("-r", "--ref_dicom", type=str, help="[.dcm] Reference DICOM file for Attribute transfer")
    parser.add_argument("-v", "--version", action="version", version="0.1.0")

    args = parser.parse_args()

    input_file = Path(args.input_file)  # TODO: add check that file is .nii/.nii.gz
    output_dir = Path(args.output_dir)  # TODO: add check that this is directory

    if not input_file.exists():
        print(f"Input file '{input_file}' not found")
        raise SystemExit(1)

    if not output_dir.exists():
        print(f"Output directory '{output_dir}' does not exist")
        raise SystemExit(1)

    # Coding of optional file checks below is quite verbose
    if args.dicom_type is not None:
        dicom_type = args.dicom_type  # TODO: add check that supplied dicom_type is permitted
    elif args.dicom_type is None:
        dicom_type = None

    if args.ref_dicom is not None:
        ref_dicom_file = Path(args.ref_dicom)   # TODO: add check that file is DICOM
    elif args.ref_dicom is None:
        ref_dicom_file = None

    # execute nii2dcm
    run_nii2dcm(
        input_file,
        output_dir,
        dicom_type,
        ref_dicom_file
    )


if __name__ == "__main__":
    sys.exit(cli())



"""
Classes for creating 3D MRI SVR DICOMs

Tom Roberts
"""

from nii2dcm.dcm import DicomMRI, nii2dcm_temp_filename


class DicomMRISVR(DicomMRI):
    """
    Creates 3D Slice-to-Volume Registration reconstruction DICOM
    """

    def __init__(self, filename=nii2dcm_temp_filename):
        super().__init__(filename)

        self.ds.MRAcquisitionType = '3D'
        self.ds.AcquisitionDuration = ''

        self.ds.ProtocolName = 'SVRTK_RESEARCH_RECONSTRUCTION'
        self.ds.SeriesDescription = 'SVRTK_RESEARCH_RECONSTRUCTION'

        """
        DICOM Attributes to transfer from DICOM supplied using --ref_dicom CLI option        
        IMPORTANT: this list overrides the equivalent list in DicomMRI class
        """

        # Not transferring PatientPosition
        # - Does this affect 3D display? Leave empty for now
        self.attributes_to_transfer.remove('PatientPosition')



"""
IOD Module – Common Instance Reference

C.12.2
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.12.2.html
"""

from nii2dcm.module import Module


class CommonInstanceReference(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'CommonInstanceReference'

        # ReferencedSeriesSequence
        # omit for now, but can be important for references Instances/Series with other Series
        # self.ds.ReferencedSeriesSequence = ''
        pass



"""
IOD Module – General Equipment

C.7.5.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.5.html#sect_C.7.5.1
"""

from nii2dcm.module import Module


class GeneralEquipment(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'GeneralEquipment'

        self.ds.Manufacturer = ''
        self.ds.InstitutionName = 'INSTITUTION_NAME_UNDEFINED'
        self.ds.ManufacturerModelName = ""
        self.ds.SoftwareVersions = ''



"""
IOD Module – SOP Common

C.12.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.12.html#sect_C.12.1
"""

from nii2dcm.module import Module


class SOPCommon(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'SOPCommon'

        self.ds.SOPClassUID = ''  # initiated, should be defined by Dicom subclass
        self.ds.SOPInstanceUID = ''
    
        # SpecificCharacterSet
        # Default set to ISO_IR 100 = Latin alphabet No. 1
        # https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.12.html#sect_C.12.1.1.2
        self.ds.SpecificCharacterSet = 'ISO_IR 100'
        self.ds.InstanceCreationDate = ''
        self.ds.InstanceCreationTime = ''
    
        self.ds.InstanceCreatorUID = ''



"""
creates a DICOM Series
"""

import os
import pydicom as pyd


def write_slice(dcm, img_data, instance_index, output_dir):
    """
    write a single DICOM slice

    dcm – nii2dcm DICOM object
    img_data - [nX, nY, nSlice] image pixel data, such as from NIfTI file
    instance_index – instance index (important: counts from 0)
    output_dir – output DICOM file save location
    """

    output_filename = r'IM_%04d' % (instance_index + 1)  # begin filename from 1, e.g. IM_0001

    img_slice = img_data[:, :, instance_index]

    # Instance UID – unique to current slice
    dcm.ds.SOPInstanceUID = pyd.uid.generate_uid(None)

    # write pixel data
    dcm.ds.PixelData = img_slice.tobytes()

    # write DICOM file
    dcm.ds.save_as(os.path.join(output_dir, output_filename), write_like_original=False)


def transfer_nii_hdr_series_tags(dcm, nii2dcm_parameters):
    """
    Transfer NIfTI header parameters applicable across Series

    dcm – nii2dcm DICOM object
    nii2dcm_parameters - parameters from NIfTI file
    """

    dcm.ds.Rows = nii2dcm_parameters['Rows']
    dcm.ds.Columns = nii2dcm_parameters['Columns']
    dcm.ds.PixelSpacing = [round(float(nii2dcm_parameters['dimX']), 2), round(float(nii2dcm_parameters['dimY']), 2)]
    dcm.ds.SliceThickness = nii2dcm_parameters['SliceThickness']
    dcm.ds.SpacingBetweenSlices = round(float(nii2dcm_parameters['SpacingBetweenSlices']), 2)
    dcm.ds.ImageOrientationPatient = nii2dcm_parameters['ImageOrientationPatient']
    dcm.ds.AcquisitionMatrix = nii2dcm_parameters['AcquisitionMatrix']
    dcm.ds.SmallestImagePixelValue = int(nii2dcm_parameters['SmallestImagePixelValue']) \
        if int(nii2dcm_parameters['SmallestImagePixelValue']) > 0 else 0  # SmallestImagePixelValue must be >= 0
    dcm.ds.LargestImagePixelValue = int(nii2dcm_parameters['LargestImagePixelValue'])
    dcm.ds.WindowCenter = nii2dcm_parameters['WindowCenter']
    dcm.ds.WindowWidth = nii2dcm_parameters['WindowWidth']
    dcm.ds.RescaleIntercept = nii2dcm_parameters['RescaleIntercept']
    dcm.ds.RescaleSlope = nii2dcm_parameters['RescaleSlope']


def transfer_nii_hdr_instance_tags(dcm, nii2dcm_parameters, instance_index):
    """
    Transfer NIfTI header parameters applicable to Instance

    dcm – nii2dcm DICOM object
    nii2dcm_parameters - parameters from NIfTI file
    instance_index - slice number in NIfTI file
    """

    # Possible per Instance Tags
    # SOPInstanceUID (set within write_slice function)
    # InstanceNumber
    # SliceLocation
    # ImagePositionPatient

    dcm.ds.InstanceNumber = nii2dcm_parameters['InstanceNumber'][instance_index]
    dcm.ds.SliceLocation = nii2dcm_parameters['SliceLocation'][instance_index]
    dcm.ds.ImagePositionPatient = [
        str(nii2dcm_parameters['ImagePositionPatient'][instance_index][0]),
        str(nii2dcm_parameters['ImagePositionPatient'][instance_index][1]),
        str(nii2dcm_parameters['ImagePositionPatient'][instance_index][2]),
    ]


def transfer_ref_dicom_series_tags(dcm, ref_dcm):
    """
    Transfer Series level DICOM Attributes from reference DICOM to nii2dcm DICOM

    dcm – nii2dcm DICOM object
    ref_dcm - reference DICOM object
    """

    for current_attribute in dcm.attributes_to_transfer:
        try:
            attribute_value = getattr(ref_dcm, current_attribute)
            setattr(dcm.ds, current_attribute, attribute_value)
        except AttributeError as e:
            # TODO once logger implemented, replace print statement with log warning
            print(f'Warning: ref_dicom {e} therefore could not transfer')



"""
IOD Module – MR Image

C.8.3.1
https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.8.3.html#sect_C.8.3.1

Note: Tags labelled ":missing:" are defined in the NEMA MR standard, but I have not found in real DICOMs exported from
an MRI scanner.
"""


from nii2dcm.module import Module


class MRImage(Module):

    def __init__(self):
        super().__init__()

        self.module_type = 'MRImage'

        # ImageType
        # NEMA defines MR-specific ImageType terms here:
        # https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.8.3.html#sect_C.8.3.1.1.1
        # For now, will omit thereby inheriting parent value
        # self.ds.ImageType = ''
    
        self.ds.SamplesPerPixel = 1
    
        # PhotometricInterpretation
        # TODO: decide MONOCHROME1 or MONOCHROME2 as default
        # https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.6.3.html#sect_C.7.6.3.1.2
        self.ds.PhotometricInterpretation = 'MONOCHROME2'
    
        # PresentationLUTShape
        # depends on PhotometricInterpretation: https://dicom.innolitics.com/ciods/mr-image/general-image/20500020
        if self.ds.PhotometricInterpretation == 'MONOCHROME2':
            self.ds.PresentationLUTShape = 'IDENTITY'
        elif self.ds.PhotometricInterpretation == 'MONOCHROME1':
            self.ds.PresentationLUTShape = 'INVERSE'
    
        # Bits Allocated
        # defined to equal 16 for MR Image Module
        # https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.8.3.html#sect_C.8.3.1.1.4
        self.ds.BitsAllocated = 16
        self.ds.BitsStored = 12
        self.ds.HighBit = self.ds.BitsStored - 1
    
        self.ds.ScanningSequence = 'RM'  # :missing:, 'RM' = Research Mode
        self.ds.SequenceVariant = ''  # :missing:
        self.ds.ScanOptions = ''  # :missing:
        self.ds.MRAcquisitionType = ''  # 2D or 3D
        self.ds.RepetitionTime = ''
        self.ds.EchoTime = ''
        self.ds.EchoTrainLength = ''
        self.ds.InversionTime = ''
        self.ds.TriggerTime = ''
        self.ds.SequenceName = ''
        self.ds.AngioFlag = ''  # :missing:
        self.ds.NumberOfAverages = ''
        self.ds.ImagingFrequency = ''
        self.ds.ImagedNucleus = ''
        self.ds.EchoNumbers = ''
        self.ds.MagneticFieldStrength = ''
        self.ds.NumberOfPhaseEncodingSteps = ''  # :missing:
        self.ds.PercentSampling = ''  # TODO set?
        self.ds.PercentPhaseFieldOfView = ''  # TODO set?
        self.ds.PixelBandwidth = ''
        self.ds.NominalInterval = ''  # :missing:
        self.ds.BeatRejectionFlag = ''  # :missing:
        self.ds.LowRRValue = ''  # :missing:
        self.ds.HighRRValue = ''  # :missing:
        self.ds.IntervalsAcquired = ''  # :missing:
        self.ds.IntervalsRejected = ''  # :missing:
        self.ds.PVCRejection = ''  # :missing:
        self.ds.SkipBeats = ''  # :missing:
        self.ds.HeartRate = ''
        self.ds.CardiacNumberOfImages = ''
        self.ds.TriggerWindow = ''
        self.ds.ReconstructionDiameter = ''  # :missing:
        self.ds.ReceiveCoilName = ''
        self.ds.TransmitCoilName = ''
        self.ds.AcquisitionMatrix = ''  # :missing:
        self.ds.InPlanePhaseEncodingDirection = ''  # ROW or COLUMN
        self.ds.FlipAngle = ''
        self.ds.SAR = ''
        self.ds.VariableFlipAngleFlag = ''  # :missing:
        self.ds.dBdt = ''
        self.ds.TemporalPositionIdentifier = ''  # :missing:
        self.ds.NumberOfTemporalPositions = ''
        self.ds.TemporalResolution = ''  # :missing:
    
        # Currently omitting, but part of NEMA MR Image module:
        # NEMA Table 10-7 “General Anatomy Optional Macro Attributes”
    
        # Currently omitting, but part of NEMA MR Image module:
        # NEMA Table 10-25 “Optional View and Slice Progression Direction Macro Attributes”
    
        self.ds.IsocenterPosition = ''  # :missing:
        self.ds.B1rms = ''




"""
Classes for creating a DICOM from scratch

Tom Roberts
"""

import datetime
import os
from random import randint

import pydicom as pyd
from pydicom.dataset import FileDataset, FileMetaDataset

from nii2dcm.utils import dcm_dictionary_update
from nii2dcm.modules.patient import Patient
from nii2dcm.modules.general_study import GeneralStudy
from nii2dcm.modules.patient_study import PatientStudy
from nii2dcm.modules.general_series import GeneralSeries
from nii2dcm.modules.frame_of_reference import FrameOfReference
from nii2dcm.modules.general_equipment import GeneralEquipment
from nii2dcm.modules.general_acquisition import GeneralAcquisition
from nii2dcm.modules.general_image import GeneralImage
from nii2dcm.modules.general_reference import GeneralReference
from nii2dcm.modules.image_plane import ImagePlane
from nii2dcm.modules.image_pixel import ImagePixel
from nii2dcm.modules.mr_image import MRImage
from nii2dcm.modules.sop_common import SOPCommon
from nii2dcm.modules.common_instance_reference import CommonInstanceReference
from nii2dcm.modules.voi_lut import VOILUT

nii2dcm_temp_filename = 'nii2dcm_tempfile.dcm'


class Dicom:
    """
    Creates basic DICOM structure

    Assumptions:
    - is_implicit_VR = False
    - is_little_endian = True
    - ImageOrientationPatient hard-coded

    """

    def __init__(self, filename=nii2dcm_temp_filename):

        self.filename = filename

        # TODO implement dcm_dictionary_update() function, most likely at this location in code
        # dcm_dictionary_update()

        # Instantiates minimal Pydicom FileMetaDataset object
        self.file_meta = FileMetaDataset()
        self.file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'  # Explicit VR Little Endian
        self.file_meta.ImplementationVersionName = 'nii2dcm_DICOM'

        # Instantiates minimal DataSet object
        self.ds = FileDataset(filename, {}, file_meta=self.file_meta, preamble=b"\0" * 128)
        self.ds.is_implicit_VR = False
        self.ds.is_little_endian = True
        self.ds.ImageType = ['DERIVED', 'SECONDARY']

        """
        Initialise Composite IOD by adding Modules to Dicom object
        """
        self.add_base_modules()

        """
        Set Dicom Date/Time
        Important: doing this once sets all Instances/Series/Study creation dates and times to the same values. Whereas, 
        doing this within the Modules would every so slightly offset the times
        """
        # TODO shift to utils.py and propagate to Modules, or, create method within this Dicom class
        dt = datetime.datetime.now()
        dateStr = dt.strftime('%Y%m%d')
        timeStr = dt.strftime('%H%M%S.%f')  # long format with micro seconds

        self.ds.ContentDate = dateStr
        self.ds.ContentTime = timeStr
        self.ds.StudyDate = dateStr
        self.ds.StudyTime = timeStr
        self.ds.SeriesDate = dateStr
        self.ds.SeriesTime = timeStr
        self.ds.AcquisitionDate = dateStr
        self.ds.AcquisitionTime = timeStr
        self.ds.InstanceCreationDate = dateStr
        self.ds.InstanceCreationTime = timeStr

        """
        Set some default Attribute values for _all_ DICOM CIODs
        """
        # TODO
        #  - decide if this is the best way to implement setting default values of important Attributes
        #  - probably makes more sense to hard-code these in subclasses, e.g. DicomMRI, DicomCT, etc, so that they are
        #  not hard-coded across different imaging modalities, where some might be irrelevant

        # ImageOrientationPatient
        # hard-coded, probably better way to define based on NIfTI values
        self.ds.ImageOrientationPatient = ['1', '0', '0', '0', '1', '0']

        # Display Attributes
        # NB: RescaleIntercept and RescaleSlope do NOT appear to be in MR Image Module, but are in CT Image, Secondary
        # Capture and Enhanced MR Modules, hence for MRI must define in this class or within DicomMRI class
        self.ds.RescaleIntercept = ''
        self.ds.RescaleSlope = ''
        self.ds.WindowCenter = ''
        self.ds.WindowWidth = ''

        # per Instance Attributes
        # initialised with hard-coded value below, but overwritten with transfer_nii_hdr_instance_tags()
        self.ds.ImagePositionPatient = ['0', '0', '0']

        self.init_study_tags()
        self.init_series_tags()

    def add_module(self, module_obj):
        """
        Update Dicom object with DICOM elements from Module object
        :param module_obj: Module object
        :return: updated Dicom object
        """

        # TODO add logger statements
        # logging(f'Updating Dicom object with DICOM elements from Module: {module.name}.')

        for elem in module_obj.ds:
            # logging('Original Element:', self.ds[elem.keyword])
            # logging('Module Element:', elem)
            self.ds.add(elem)
            # logging('New Tag:', self.ds[elem.keyword])

    def add_base_modules(self):
        """
        Initialise Composite IOD by adding Modules to Dicom object

        Common Modules below taken by comparing A.3.3 (CT) and A.4.3 (MRI) CIODs to determine shared Modules. Other
        modalities assumed to have similar composition. Modules unique to a specific imaging modality are added within
        the respective subclass. e.g. MR Image Module is added within the nii2dcm DicomMRI class
        """
        self.add_module(Patient())
        self.add_module(GeneralStudy())
        self.add_module(PatientStudy())
        self.add_module(GeneralSeries())
        self.add_module(FrameOfReference())
        self.add_module(GeneralEquipment())

        # Image IE
        # TODO potentially transfer these Modules to Dicom subclasses because they are image-based,
        #  whereas Modules above are generic, non-image Modules, e.g. patient details, file information etc.
        self.add_module(GeneralImage())
        self.add_module(GeneralAcquisition())
        self.add_module(GeneralReference())
        self.add_module(ImagePlane())
        self.add_module(ImagePixel())
        self.add_module(SOPCommon())
        self.add_module(CommonInstanceReference())
        self.add_module(VOILUT())

    def get_file_meta(self):
        return self.file_meta

    def get_dataset(self):
        return self.ds

    def save_as(self):
        print("Writing DICOM to", os.path.join(os.getcwd(), self.filename))
        self.ds.save_as(self.filename)

    def init_study_tags(self):
        """
        Create Study Tags
        - these tags are fixed across Instances and Series
        """

        # Possible per Study Tags
        # StudyInstanceUID

        self.ds.StudyInstanceUID = pyd.uid.generate_uid(None)

    def init_series_tags(self):
        """
        Create Series Tags
        - these tags are fixed across Instances
        """

        # Possible per Series Tags
        # SeriesInstanceUID
        # FrameOfReferenceUID
        # SeriesDate
        # SeriesTime
        # AcquisitionDate
        # AcquisitionTime
        # SeriesNumber

        self.ds.SeriesInstanceUID = pyd.uid.generate_uid(None)
        self.ds.FrameOfReferenceUID = pyd.uid.generate_uid(None)
        self.ds.SeriesNumber = str(randint(1000, 9999))  # 4 digit number to avoid conflict
        self.ds.AcquisitionNumber = ''  # Innolitics says this is in General Image Module, but NEMA says it is not...


class DicomMRI(Dicom):
    """
    DicomMRI subclass
    - Sets appropriate SOPClass UIDs for MR
    - Adds MR Image Module to Dicom object
    """

    def __init__(self, filename=nii2dcm_temp_filename):
        super().__init__(filename)

        """
        Set DICOM attributes which are located outside of the MR Image Module to MR-specific values 
        """
        self.ds.Modality = 'MR'

        # MR Image Storage SOP Class
        # UID = 1.2.840.10008.5.1.4.1.1.4
        # https://dicom.nema.org/dicom/2013/output/chtml/part04/sect_I.4.html
        self.file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.4'
        self.ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.4'

        """
        Initialise subclass CIOD Modules
        """
        self.add_module(MRImage())

        """
        DICOM Attributes to transfer from DICOM supplied using --ref_dicom CLI option
        """
        # TODO(tomaroberts) figure out whether transfer SQ Sequence blocks, e.g.:
        #  ProcedureCodeSequence,
        #  ReferencedStudySequence,
        #  ConversionSourceAttributesSequence
        #  - conflicts if _included_?

        self.attributes_to_transfer = [
            'StudyDate',
            'SeriesDate',  # TODO omit so SeriesDate = when DICOM created, not acquired? Consistent with SeriesTime
            'AcquisitionDate',
            'AccessionNumber',
            'InstitutionName',
            'InstitutionAddress',
            'ReferringPhysicianName',
            'StationName',
            'StudyDescription',
            'ProcedureCodeSequence',  # SQ Sequence
            'InstitutionalDepartmentName',
            'PerformingPhysicianName',
            'OperatorsName',
            'ManufacturerModelName',
            'ReferencedStudySequence',  # SQ Sequence
            'RelatedSeriesSequence',  # SQ Sequence

            'PatientName',
            'PatientID',
            'PatientBirthDate',
            'PatientSex',
            'PatientAge',
            'PatientSize',
            'PatientWeight',
            'BodyPartExamined',

            # MR Image Module Attributes
            'ScanningSequence',
            'SequenceVariant',
            'ScanOptions',
            'MRAcquisitionType',
            'SequenceName',
            'AngioFlag',
            'RepetitionTime',
            'EchoTime',
            'InversionTime',
            'NumberOfAverages',
            'ImagingFrequency',
            'ImagedNucleus',
            'MagneticFieldStrength',
            'NumberOfPhaseEncodingSteps',
            'EchoTrainLength',
            'PercentSampling',
            'PercentPhaseFieldOfView',
            'PixelBandwidth',
            'DeviceSerialNumber',
            'SoftwareVersions',
            'BeatRejectionFlag',
            'CardiacNumberOfImages',
            'ReceiveCoilName',
            'TransmitCoilName',
            'InPlanePhaseEncodingDirection',
            'FlipAngle',
            'SAR',
            'PatientPosition',

            # General Study Module Attributes
            'StudyInstanceUID',  # Important: enables new DICOM to be filed in original Study
            'StudyID',
            'AcquisitionNumber',  # include or set in Dicom subclass?
            'FrameOfReferenceUID',  # include?
            'NumberOfTemporalPositions',
            'ConversionSourceAttributesSequence',  # SQ Sequence
            'RequestAttributesSequence',  # SQ Sequence
            'RequestingPhysician',
            'RequestingService',
        ]





from setuptools import setup, find_packages
from pathlib import Path


# import README.md for long_description
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="nii2dcm",
    version="0.1.1",
    description='nii2dcm: NIfTI to DICOM creation with Python',
    author='Tom Roberts',
    url='https://github.com/tomaroberts/nii2dcm',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(include=['nii2dcm', 'nii2dcm.*']),
    install_requires=[
        'numpy',
        'matplotlib',
        'nibabel',
        'pydicom'
    ],
    setup_requires=['flake8'],
    entry_points={
        "console_scripts": [
            "nii2dcm = nii2dcm.__main__:cli"
        ]
    },
)